@@ -1,7 +1,8 @@
 (*
  * curl.ml
  *
- * Copyright (c) 2003, Lars Nilsson, <lars@quantumchamaeleon.com>
+ * Copyright (c) 2003-2008, Lars Nilsson, <lars@quantumchamaeleon.com>
+ * Copyright (c) 2009, ygrek, <ygrek@autistici.org>
  *)
 
 type t
@@ -71,7 +72,24 @@ type curlCode =
   | CURLE_BAD_CONTENT_ENCODING
   | CURLE_LDAP_INVALID_URL
   | CURLE_FILESIZE_EXCEEDED
-  | CURLE_FTP_SSL_FAILED
+  | CURLE_USE_SSL_FAILED
+  | CURLE_SEND_FAIL_REWIND
+  | CURLE_SSL_ENGINE_INITFAILED
+  | CURLE_LOGIN_DENIED
+  | CURLE_TFTP_NOTFOUND
+  | CURLE_TFTP_PERM
+  | CURLE_REMOTE_DISK_FULL
+  | CURLE_TFTP_ILLEGAL
+  | CURLE_TFTP_UNKNOWNID
+  | CURLE_REMOTE_FILE_EXISTS
+  | CURLE_TFTP_NOSUCHUSER
+  | CURLE_CONV_FAILED
+  | CURLE_CONV_REQD
+  | CURLE_SSL_CACERT_BADFILE
+  | CURLE_REMOTE_FILE_NOT_FOUND
+  | CURLE_SSH
+  | CURLE_SSL_SHUTDOWN_FAILED
+  | CURLE_AGAIN
 
 exception CurlException of (curlCode * int * string)
 
@@ -83,7 +101,9 @@ type curlNETRCOption =
 type curlEncoding =
   | CURL_ENCODING_NONE
   | CURL_ENCODING_DEFLATE
-      
+  | CURL_ENCODING_GZIP
+  | CURL_ENCODING_ANY
+
 type curlContentType =
   | DEFAULT
   | CONTENTTYPE of string
@@ -94,7 +114,7 @@ type curlHTTPPost =
   | CURLFORM_FILE of string * string * curlContentType
   | CURLFORM_BUFFER of string * string * string * curlContentType
 
-(*      
+(*
 type curlHTTPPost =
   | CURLFORM_COPYNAME of string
   | CURLFORM_PTRNAME of string
@@ -108,8 +128,10 @@ type curlHTTPPost =
 *)
 
 type curlTimeCondition =
+  | TIMECOND_NONE
   | TIMECOND_IFMODSINCE
   | TIMECOND_IFUNMODSINCE
+  | TIMECOND_LASTMOD
 
 type curlKRB4Level =
   | KRB4_NONE
@@ -123,6 +145,7 @@ type curlClosePolicy =
   | CLOSEPOLICY_LEAST_RECENTLY_USED
 
 type curlSSLVerifyHost =
+  | SSLVERIFYHOST_NONE
   | SSLVERIFYHOST_EXISTENCE
   | SSLVERIFYHOST_HOSTNAME
 
@@ -158,8 +181,88 @@ type curlFTPSSL =
   | FTPSSL_CONTROL
   | FTPSSL_ALL
 
+type curlFTPSSLAuth =
+  | FTPAUTH_DEFAULT
+  | FTPAUTH_SSL
+  | FTPAUTH_TLS
+
+type curlIOCmd =
+  | IOCMD_NOP
+  | IOCMD_RESTARTREAD
+
+type curlIOErr =
+  | IOE_OK
+  | IOE_UNKNOWNCMD
+  | IOE_FAILRESTART
+
+type curlSeekResult =
+  | SEEKFUNC_OK
+  | SEEKFUNC_FAIL
+  | SEEKFUNC_CANTSEEK
+
+type curlFTPMethod =
+  | FTPMETHOD_DEFAULT
+  | FTPMETHOD_MULTICWD
+  | FTPMETHOD_NOCWD
+  | FTPMETHOD_SINGLECWD
+
+type curlSSHAuthTypes =
+  | SSHAUTH_ANY
+  | SSHAUTH_PUBLICKEY
+  | SSHAUTH_PASSWORD
+  | SSHAUTH_HOST
+  | SSHAUTH_KEYBOARD
+
+type curlFTPSSLCCC =
+  | FTPSSL_CCC_NONE
+  | FTPSSL_CCC_PASSIVE
+  | FTPSSL_CCC_ACTIVE
+
+type curlSeek =
+  | SEEK_SET
+  | SEEK_CUR
+  | SEEK_END
+
+type curlProxyType =
+  | CURLPROXY_HTTP
+  | CURLPROXY_HTTP_1_0 (** added in 7.19.4 *)
+  | CURLPROXY_SOCKS4 (** added in 7.15.2 *)
+  | CURLPROXY_SOCKS5
+  | CURLPROXY_SOCKS4A (** added in 7.18.0 *)
+  | CURLPROXY_SOCKS5_HOSTNAME (** added in 7.18.0 *)
+
+(** Protocols to enable (via CURLOPT_PROTOCOLS and CURLOPT_REDIR_PROTOCOLS) *)
+type curlProto =
+| CURLPROTO_ALL (** enable everything *)
+| CURLPROTO_HTTP
+| CURLPROTO_HTTPS
+| CURLPROTO_FTP
+| CURLPROTO_FTPS
+| CURLPROTO_SCP
+| CURLPROTO_SFTP
+| CURLPROTO_TELNET
+| CURLPROTO_LDAP
+| CURLPROTO_LDAPS
+| CURLPROTO_DICT
+| CURLPROTO_FILE
+| CURLPROTO_TFTP
+| CURLPROTO_IMAP
+| CURLPROTO_IMAPS
+| CURLPROTO_POP3
+| CURLPROTO_POP3S
+| CURLPROTO_SMTP
+| CURLPROTO_SMTPS
+| CURLPROTO_RTSP
+| CURLPROTO_RTMP
+| CURLPROTO_RTMPT
+| CURLPROTO_RTMPE
+| CURLPROTO_RTMPTE
+| CURLPROTO_RTMPS
+| CURLPROTO_RTMPTS
+| CURLPROTO_GOPHER
+
 type curlOption =
-  | CURLOPT_WRITEFUNCTION of (string -> unit)
+  | CURLOPT_WRITEFUNCTION of (string -> int)
   | CURLOPT_READFUNCTION of (int -> string)
   | CURLOPT_INFILESIZE of int
   | CURLOPT_URL of string
@@ -208,13 +311,12 @@ type curlOption =
   | CURLOPT_CRLF of bool
   | CURLOPT_QUOTE of string list
   | CURLOPT_POSTQUOTE of string list
-  | CURLOPT_HEADERFUNCTION of (string -> unit)
+  | CURLOPT_HEADERFUNCTION of (string -> int)
   | CURLOPT_COOKIEFILE of string
   | CURLOPT_SSLVERSION of int
   | CURLOPT_TIMECONDITION of curlTimeCondition
   | CURLOPT_TIMEVALUE of int32
   | CURLOPT_CUSTOMREQUEST of string
-  | CURLOPT_STDERR (* UNIMPLEMENTED *)
   | CURLOPT_INTERFACE of string
   | CURLOPT_KRB4LEVEL of curlKRB4Level
   | CURLOPT_PROGRESSFUNCTION of (float -> float -> float -> float -> bool)
@@ -256,6 +358,44 @@ type curlOption =
   | CURLOPT_FTPSSL of curlFTPSSL
   | CURLOPT_POSTFIELDSIZELARGE of int64
   | CURLOPT_TCPNODELAY of bool
+  | CURLOPT_FTPSSLAUTH of curlFTPSSLAuth
+  | CURLOPT_IOCTLFUNCTION of (t -> curlIOCmd -> curlIOErr)
+  | CURLOPT_FTPACCOUNT of string
+  | CURLOPT_COOKIELIST of string
+  | CURLOPT_IGNORECONTENTLENGTH of bool
+  | CURLOPT_FTPSKIPPASVIP of bool
+  | CURLOPT_FTPFILEMETHOD of curlFTPMethod
+  | CURLOPT_LOCALPORT of int
+  | CURLOPT_LOCALPORTRANGE of int
+  | CURLOPT_CONNECTONLY of bool
+  | CURLOPT_MAXSENDSPEEDLARGE of int64
+  | CURLOPT_MAXRECVSPEEDLARGE of int64
+  | CURLOPT_FTPALTERNATIVETOUSER of string
+  | CURLOPT_SSLSESSIONIDCACHE of bool
+  | CURLOPT_SSHAUTHTYPES of curlSSHAuthTypes list
+  | CURLOPT_SSHPUBLICKEYFILE of string
+  | CURLOPT_SSHPRIVATEKEYFILE of string
+  | CURLOPT_FTPSSLCCC of curlFTPSSLCCC
+  | CURLOPT_TIMEOUTMS of int
+  | CURLOPT_CONNECTTIMEOUTMS of int
+  | CURLOPT_HTTPTRANSFERDECODING of bool
+  | CURLOPT_HTTPCONTENTDECODING of bool
+  | CURLOPT_NEWFILEPERMS of int
+  | CURLOPT_NEWDIRECTORYPERMS of int
+  | CURLOPT_POST301 of bool
+  | CURLOPT_SSHHOSTPUBLICKEYMD5 of string
+  | CURLOPT_COPYPOSTFIELDS of string
+  | CURLOPT_PROXYTRANSFERMODE of bool
+  | CURLOPT_SEEKFUNCTION of (int64 -> curlSeek -> curlSeekResult)
+  | CURLOPT_AUTOREFERER of bool
+  | CURLOPT_OPENSOCKETFUNCTION of (Unix.file_descr -> unit)
+  | CURLOPT_PROXYTYPE of curlProxyType
+  | CURLOPT_PROTOCOLS of curlProto list
+  | CURLOPT_REDIR_PROTOCOLS of curlProto list
+  | CURLOPT_RESOLVE of string list
+  | CURLOPT_DNS_SERVERS of string
+  | CURLOPT_MAIL_FROM of string
+  | CURLOPT_MAIL_RCPT of string list
 
 type initOption =
   | CURLINIT_GLOBALALL
@@ -289,11 +429,27 @@ type curlInfo =
   | CURLINFO_HTTP_CONNECTCODE
   | CURLINFO_HTTPAUTH_AVAIL
   | CURLINFO_PROXYAUTH_AVAIL
+  | CURLINFO_OS_ERRNO
+  | CURLINFO_NUM_CONNECTS
+  | CURLINFO_SSL_ENGINES
+  | CURLINFO_COOKIELIST
+  | CURLINFO_LASTSOCKET
+  | CURLINFO_FTP_ENTRY_PATH
+  | CURLINFO_REDIRECT_URL
+  | CURLINFO_PRIMARY_IP
+  | CURLINFO_LOCAL_IP
+  | CURLINFO_LOCAL_PORT
+  | CURLINFO_CONDITION_UNMET
 
 type curlInfoResult =
   | CURLINFO_String of string
   | CURLINFO_Long of int
   | CURLINFO_Double of float
+  | CURLINFO_StringList of string list
+
+exception NotImplemented of string
+
+let () = Callback.register_exception "Curl.NotImplemented" (NotImplemented "")
 
 external helper_global_init : initOption -> unit = "helper_curl_global_init"
 external helper_global_cleanup : unit -> unit = "helper_curl_global_cleanup"
@@ -307,10 +463,12 @@ external helper_escape : string -> string = "helper_curl_escape"
 external helper_unescape : string -> string = "helper_curl_unescape"
 external helper_getdate : string -> float -> float = "helper_curl_getdate"
 external helper_version : unit -> string = "helper_curl_version"
-					     
+
 let init () =
   helper_init ()
 
+external reset : t -> unit = "helper_curl_easy_reset"
+
 let duphandle conn =
   helper_duphandle conn
 
@@ -344,6 +502,30 @@ let getdate str now =
 let version () =
   helper_version ()
 
+type version_info = { 
+  version : string;
+  number : int * int * int;
+  host : string;
+  features : string list;
+  ssl_version : string option;
+  libz_version : string option;
+  protocols : string list;
+  ares : string;
+  ares_num : int;
+  libidn : string;
+  iconv_ver_num : int;
+  libssh_version : string;
+}
+
+external version_info : unit -> version_info = "caml_curl_version_info"
+
+external strerror : curlCode -> string = "helper_curl_easy_strerror"
+let errno : curlCode -> int = Obj.magic
+
+type pauseOption = PAUSE_SEND | PAUSE_RECV | PAUSE_ALL
+
+external pause : t -> pauseOption list -> unit = "caml_curl_pause"
+
 let set_writefunction conn closure =
   setopt conn (CURLOPT_WRITEFUNCTION closure)
 
@@ -632,118 +814,239 @@ let set_postfieldsizelarge conn size =
 let set_tcpnodelay conn flag =
   setopt conn (CURLOPT_TCPNODELAY flag)
 
+let set_ftpsslauth conn auth =
+  setopt conn (CURLOPT_FTPSSLAUTH auth)
+
+let set_ioctlfunction conn closure =
+  setopt conn (CURLOPT_IOCTLFUNCTION closure)
+
+let set_ftpaccount conn account =
+  setopt conn (CURLOPT_FTPACCOUNT account)
+
+let set_cookielist conn cookielist =
+  setopt conn (CURLOPT_COOKIELIST cookielist)
+
+let set_ignorecontentlength conn flag =
+  setopt conn (CURLOPT_IGNORECONTENTLENGTH flag)
+
+let set_ftpskippasvip conn flag =
+  setopt conn (CURLOPT_FTPSKIPPASVIP flag)
+
+let set_ftpfilemethod conn ftpMethod =
+  setopt conn (CURLOPT_FTPFILEMETHOD ftpMethod)
+
+let set_localport conn port =
+  setopt conn (CURLOPT_LOCALPORT port)
+
+let set_localportrange conn range =
+  setopt conn (CURLOPT_LOCALPORTRANGE range)
+
+let set_connectonly conn flag =
+  setopt conn (CURLOPT_CONNECTONLY flag)
+
+let set_maxsendspeedlarge conn speed =
+  setopt conn (CURLOPT_MAXSENDSPEEDLARGE speed)
+
+let set_maxrecvspeedlarge conn speed =
+  setopt conn (CURLOPT_MAXRECVSPEEDLARGE speed)
+
+let set_ftpalternativetouser conn user =
+  setopt conn (CURLOPT_FTPALTERNATIVETOUSER user)
+
+let set_sslsessionidcache conn flag =
+  setopt conn (CURLOPT_SSLSESSIONIDCACHE flag)
+
+let set_sshauthtypes conn types =
+  setopt conn (CURLOPT_SSHAUTHTYPES types)
+
+let set_sshpublickeyfile conn keyfile =
+  setopt conn (CURLOPT_SSHPUBLICKEYFILE keyfile)
+
+let set_sshprivatekeyfile conn keyfile =
+  setopt conn (CURLOPT_SSHPRIVATEKEYFILE keyfile)
+
+let set_ftpsslccc conn ccc =
+  setopt conn (CURLOPT_FTPSSLCCC ccc)
+
+let set_timeoutms conn ms =
+  setopt conn (CURLOPT_TIMEOUTMS ms)
+
+let set_connecttimeoutms conn ms =
+  setopt conn (CURLOPT_CONNECTTIMEOUTMS ms)
+
+let set_httptransferdecoding conn flag =
+  setopt conn (CURLOPT_HTTPTRANSFERDECODING flag)
+
+let set_httpcontentdecoding conn flag =
+  setopt conn (CURLOPT_HTTPCONTENTDECODING flag)
+
+let set_newfileperms conn perms =
+  setopt conn (CURLOPT_NEWFILEPERMS perms)
+
+let set_newdirectoryperms conn perms =
+  setopt conn (CURLOPT_NEWDIRECTORYPERMS perms)
+
+let set_post301 conn flag =
+  setopt conn (CURLOPT_POST301 flag)
+
+let set_sshhostpublickeymd5 conn key =
+  setopt conn (CURLOPT_SSHHOSTPUBLICKEYMD5 key)
+
+let set_copypostfields conn post =
+  setopt conn (CURLOPT_COPYPOSTFIELDS post)
+
+let set_proxytransfermode conn flag =
+  setopt conn (CURLOPT_PROXYTRANSFERMODE flag)
+
+let set_seekfunction conn closure =
+  setopt conn (CURLOPT_SEEKFUNCTION closure)
+
+let set_autoreferer conn b =
+  setopt conn (CURLOPT_AUTOREFERER b)
+
+let set_opensocketfunction conn closure =
+  setopt conn (CURLOPT_OPENSOCKETFUNCTION closure)
+
+let set_proxytype conn ptype =
+  setopt conn (CURLOPT_PROXYTYPE ptype)
+
+let set_protocols conn l =
+  setopt conn (CURLOPT_PROTOCOLS l)
+
+let set_redirprotocols conn l =
+  setopt conn (CURLOPT_REDIR_PROTOCOLS l)
+
+let set_resolve conn l_add l_del =
+  let acc = List.fold_left (fun acc (host,port,address) -> (host ^ ":" ^ string_of_int port ^ ":" ^ address) :: acc) [] l_add in
+  let acc = List.fold_left (fun acc (host,port) -> ("-" ^ host ^ ":" ^ string_of_int port) :: acc) acc l_del in
+  setopt conn (CURLOPT_RESOLVE acc)
+
+let set_dns_servers conn l =
+  setopt conn (CURLOPT_DNS_SERVERS (String.concat "," l))
+
+let set_mailfrom conn l =
+  setopt conn (CURLOPT_MAIL_FROM l)
+
+let set_mailrcpt conn l =
+  setopt conn (CURLOPT_MAIL_RCPT l)
+
 let get_effectiveurl conn =
   match (getinfo conn CURLINFO_EFFECTIVE_URL) with
-      CURLINFO_String s -> s
-    | _ -> ""
+  | CURLINFO_String s -> s
+  | _ -> ""
+
+let get_redirecturl conn =
+  match (getinfo conn CURLINFO_REDIRECT_URL) with
+  | CURLINFO_String s -> s
+  | _ -> ""
 
 let get_responsecode conn =
   match (getinfo conn CURLINFO_HTTP_CODE) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  | CURLINFO_Long l -> l
+  | _ -> 0
 
 let get_httpcode conn =
   get_responsecode conn
 
 let get_totaltime conn =
   match (getinfo conn CURLINFO_TOTAL_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  | CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_namelookuptime conn =
   match (getinfo conn CURLINFO_NAMELOOKUP_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  | CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_connecttime conn =
   match (getinfo conn CURLINFO_CONNECT_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  | CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_pretransfertime conn =
   match (getinfo conn CURLINFO_PRETRANSFER_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_sizeupload conn =
   match (getinfo conn CURLINFO_SIZE_UPLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_sizedownload conn =
   match (getinfo conn CURLINFO_SIZE_DOWNLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_speeddownload conn =
   match (getinfo conn CURLINFO_SPEED_DOWNLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_speedupload conn =
   match (getinfo conn CURLINFO_SPEED_UPLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_headersize conn =
   match (getinfo conn CURLINFO_HEADER_SIZE) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  |  CURLINFO_Long l -> l
+  | _ -> 0
 
 let get_requestsize conn =
   match (getinfo conn CURLINFO_REQUEST_SIZE) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  |  CURLINFO_Long l -> l
+  | _ -> 0
 
 let get_sslverifyresult conn =
   match (getinfo conn CURLINFO_SSL_VERIFYRESULT) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  |  CURLINFO_Long l -> l
+  | _ -> 0
 
 let get_filetime conn =
   match (getinfo conn CURLINFO_FILETIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_contentlengthdownload conn =
   match (getinfo conn CURLINFO_CONTENT_LENGTH_DOWNLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_contentlengthupload conn =
   match (getinfo conn CURLINFO_CONTENT_LENGTH_UPLOAD) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_starttransfertime conn =
   match (getinfo conn CURLINFO_STARTTRANSFER_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_contenttype conn =
   match (getinfo conn CURLINFO_CONTENT_TYPE) with
-      CURLINFO_String s -> s
-    | _ -> ""
+  |  CURLINFO_String s -> s
+  | _ -> ""
 
 let get_redirecttime conn =
   match (getinfo conn CURLINFO_REDIRECT_TIME) with
-      CURLINFO_Double d -> d
-    | _ -> 0.0
+  |  CURLINFO_Double d -> d
+  | _ -> 0.0
 
 let get_redirectcount conn =
   match (getinfo conn CURLINFO_REDIRECT_COUNT) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  |  CURLINFO_Long l -> l
+  | _ -> 0
 
 let get_private conn =
   match (getinfo conn CURLINFO_PRIVATE) with
-      CURLINFO_String s -> s
-    | _ -> ""
+  |  CURLINFO_String s -> s
+  | _ -> ""
 
 let get_httpconnectcode conn =
   match (getinfo conn CURLINFO_HTTP_CONNECTCODE) with
-      CURLINFO_Long l -> l
-    | _ -> 0
+  |  CURLINFO_Long l -> l
+  | _ -> 0
 
 let generate_auth auth =
   let result = ref [] in
@@ -755,13 +1058,63 @@ let generate_auth auth =
 
 let get_httpauthavail conn =
   match (getinfo conn CURLINFO_HTTPAUTH_AVAIL) with
-      CURLINFO_Long l -> generate_auth l
-    | _ -> []
+  | CURLINFO_Long l -> generate_auth l
+  | _ -> []
 
 let get_proxyauthavail conn =
   match (getinfo conn CURLINFO_PROXYAUTH_AVAIL) with
-      CURLINFO_Long l -> generate_auth l
-    | _ -> []
+  | CURLINFO_Long l -> generate_auth l
+  | _ -> []
+
+let get_oserrno conn =
+  match (getinfo conn CURLINFO_OS_ERRNO) with
+  | CURLINFO_Long l -> l
+  | _ -> 0
+
+let get_numconnects conn =
+  match (getinfo conn CURLINFO_NUM_CONNECTS) with
+  | CURLINFO_Long l -> l
+  | _ -> 0
+
+let get_sslengines conn =
+  match (getinfo conn CURLINFO_SSL_ENGINES) with
+  | CURLINFO_StringList l -> l
+  | _ -> []
+
+let get_cookielist conn =
+  match (getinfo conn CURLINFO_COOKIELIST) with
+  | CURLINFO_StringList l -> l
+  | _ -> []
+
+let get_lastsocket conn =
+  match (getinfo conn CURLINFO_LASTSOCKET) with
+  | CURLINFO_Long l -> l
+  | _ -> 0
+
+let get_ftpentrypath conn =
+  match (getinfo conn CURLINFO_FTP_ENTRY_PATH) with
+  | CURLINFO_String s -> s
+  | _ -> ""
+
+let get_primaryip conn =
+  match (getinfo conn CURLINFO_PRIMARY_IP) with
+  | CURLINFO_String s -> s
+  | _ -> ""
+
+let get_localip conn =
+  match (getinfo conn CURLINFO_LOCAL_IP) with
+  | CURLINFO_String s -> s
+  | _ -> ""
+
+let get_localport conn =
+  match (getinfo conn CURLINFO_LOCAL_PORT) with
+  | CURLINFO_Long n -> n
+  | _ -> 0
+
+let get_conditionunmet conn =
+  match (getinfo conn CURLINFO_CONDITION_UNMET) with
+  | CURLINFO_Long n -> n <> 0
+  | _ -> assert false
 
 let () =
   Callback.register_exception "CurlException"
@@ -770,6 +1123,7 @@ let () =
 class handle =
   object
     val conn = init ()
+    method handle = conn
     method duphandle = {< conn = duphandle conn >}
     method perform = perform conn
     method cleanup = cleanup conn
@@ -870,8 +1224,43 @@ class handle =
     method set_postfieldsize size = set_postfieldsize conn size
     method set_postfieldsizelarge size = set_postfieldsizelarge conn size
     method set_tcpnodelay flag = set_tcpnodelay conn flag
-					 
+    method set_ftpsslauth flag = set_ftpsslauth conn flag
+    method set_ioctlfunction closure = set_ioctlfunction conn closure
+    method set_ftpaccount account = set_ftpaccount conn account
+    method set_cookielist cookielist = set_cookielist conn cookielist
+    method set_ignorecontentlength flag = set_ignorecontentlength conn flag
+    method set_ftpskippasvip flag = set_ftpskippasvip conn flag
+    method set_ftpfilemethod ftpMethod = set_ftpfilemethod conn ftpMethod
+    method set_localport port = set_localport conn port
+    method set_localportrange range = set_localportrange conn range
+    method set_connectonly flag = set_connectonly conn flag
+    method set_maxsendspeedlarge speed = set_maxsendspeedlarge conn speed
+    method set_maxrecvspeedlarge speed = set_maxrecvspeedlarge conn speed
+    method set_ftpalternativetouser user = set_ftpalternativetouser conn user
+    method set_sslsessionidcache flag = set_sslsessionidcache conn flag
+    method set_sshauthtypes types = set_sshauthtypes conn types
+    method set_sshpublickeyfile keyfile = set_sshpublickeyfile conn keyfile
+    method set_sshprivatekeyfile keyfile = set_sshprivatekeyfile conn keyfile
+    method set_ftpsslccc ccc = set_ftpsslccc conn ccc
+    method set_timeoutms ms = set_timeoutms conn ms
+    method set_connecttimeoutms ms = set_connecttimeoutms conn ms
+    method set_httptransferdecoding flag = set_httptransferdecoding conn flag
+    method set_httpcontentdecoding flag = set_httpcontentdecoding conn flag
+    method set_newfileperms perms = set_newfileperms conn perms
+    method set_newdirectoryperms perms = set_newdirectoryperms conn perms
+    method set_post301 flag = set_post301 conn flag
+    method set_sshhostpublickeymd5 key = set_sshhostpublickeymd5 conn key
+    method set_copypostfields post = set_copypostfields conn post
+    method set_proxytransfermode flag = set_proxytransfermode conn flag
+    method set_seekfunction closure = set_seekfunction conn closure
+    method set_resolve l = set_resolve conn l
+    method set_dns_servers l = set_dns_servers conn l
+    method set_autoreferer b = set_autoreferer conn b
+    method set_opensocketfunction closure = set_opensocketfunction conn closure
+    method set_proxytype t = set_proxytype conn t
+
     method get_effectiveurl = get_effectiveurl conn
+    method get_redirecturl = get_redirecturl conn
     method get_responsecode = get_responsecode conn
     method get_httpcode = get_responsecode conn
     method get_totaltime = get_totaltime conn
@@ -896,4 +1285,49 @@ class handle =
     method get_httpconnectcode = get_httpconnectcode conn
     method get_httpauthavail = get_httpauthavail conn
     method get_proxyauthavail = get_proxyauthavail conn
+    method get_oserrno = get_oserrno conn
+    method get_numconnects = get_numconnects conn
+    method get_sslengines = get_sslengines conn
+    method get_cookielist = get_cookielist conn
+    method get_lastsocket = get_lastsocket conn
+    method get_ftpentrypath = get_ftpentrypath conn
+    method get_primaryip = get_primaryip conn
+    method get_localip = get_localip conn
+    method get_localport = get_localport conn
+    method get_conditionunmet = get_conditionunmet conn
 end
+
+module Multi = struct
+
+  type mt
+
+  exception Error of string
+
+  let () = Callback.register_exception "Curl.Multi.Error" (Error "")
+
+  external create : unit -> mt = "caml_curl_multi_init"
+  external add : mt -> t -> unit = "caml_curl_multi_add_handle"
+  external perform : mt -> int = "caml_curl_multi_perform_all"
+  external wait : mt -> bool = "caml_curlm_wait_data"
+  external remove : mt -> t -> unit = "caml_curl_multi_remove_handle"
+  external remove_finished : mt -> (t * curlCode) option = "caml_curlm_remove_finished"
+  external cleanup : mt -> unit = "caml_curl_multi_cleanup"
+
+  (* see curlm_sock_cb *)
+  type poll = POLL_NONE | POLL_IN | POLL_OUT | POLL_INOUT | POLL_REMOVE
+
+  (* see caml_curl_multi_socket_action *)
+  type fd_status = EV_AUTO | EV_IN | EV_OUT | EV_INOUT
+
+  external set_socket_function : mt -> (Unix.file_descr -> poll -> unit) -> unit = "caml_curl_multi_socketfunction"
+  external set_timer_function : mt -> (int -> unit) -> unit = "caml_curl_multi_timerfunction"
+  external action_all : mt -> int = "caml_curl_multi_socket_all"
+  external socket_action : mt -> Unix.file_descr option -> fd_status -> int = "caml_curl_multi_socket_action"
+
+  let action_timeout mt = ignore (socket_action mt None EV_AUTO)
+  let action mt fd status = socket_action mt (Some fd) status
+
+  external timeout : mt -> int = "caml_curl_multi_timeout"
+
+end
+
