@@ -1,7 +1,7 @@
 (*
  * ominimal.ml
  *
- * Copyright (c) 2003, Lars Nilsson, <lars@quantumchamaeleon.com>
+ * Copyright (c) 2003-2008, Lars Nilsson, <lars@quantumchamaeleon.com>
  *)
 
 let _ =
@@ -10,7 +10,7 @@ let _ =
       Curl.global_init Curl.CURLINIT_GLOBALALL;
       let connection = new Curl.handle in
 	connection#set_url Sys.argv.(1);
-	connection#perform
+	connection#perform;
     end
   else
     Printf.fprintf stderr "Usage: %s <url>\n" Sys.argv.(0)
