@@ -1,13 +1,13 @@
 (*
  * oput.ml
  *
- * Copyright (c) 2003, Lars Nilsson, <lars@quantumchamaeleon.com>
+ * Copyright (c) 2003-2008, Lars Nilsson, <lars@quantumchamaeleon.com>
  *)
 
 let counter = ref 0
 
 let reader file maxBytes =
-  let buffer = String.create maxBytes in
+  let buffer = String.make maxBytes ' ' in
   let readBytes = input file buffer 0 maxBytes in
     if readBytes = 0 then ""
     else
